abcdefghijklmnop.in0   abcdefghijklmnop.f00    0.0e+0    0.0e+0    00    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f01    1.0e-5    0.0e+0    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f02    0.0e+0    1.0e-5    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f03   -1.0e-5    1.0e-5    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f04    1.0e-5   -1.0e-5    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f05    1.0e-5    1.0e-5    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f06    1.0e-4    0.0e+0    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f07    0.0e+0    1.0e-4    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f08   -1.0e-4    1.0e-4    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f09    1.0e-4   -1.0e-4    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f10    1.0e-4    1.0e-4    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f11    1.0e-3    0.0e+0    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f12    0.0e+0    1.0e-3    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f13   -1.0e-3    1.0e-3    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f14    1.0e-3   -1.0e-3    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f15    1.0e-3    1.0e-3    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f16    1.0e-2    0.0e+0    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f17    0.0e+0    1.0e-2    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f18   -1.0e-2    1.0e-2    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f19    1.0e-2   -1.0e-2    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f20    1.0e-2    1.0e-2    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f21    1.0e-1    0.0e+0    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f22    0.0e+0    1.0e-1    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f23   -1.0e-1    1.0e-1    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f24    1.0e-1   -1.0e-1    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f25    1.0e-1    1.0e-1    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f26    1.0e+0    0.0e+0    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f27    0.0e+0    1.0e+0    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f28   -1.0e+0    1.0e+0    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f29    1.0e+0   -1.0e+0    25    454589
abcdefghijklmnop.in0   abcdefghijklmnop.f30    1.0e+0    1.0e+0    25    454589

a = Lags in VAR part, Lu, int
b = Lags in GARCH (autoregressive) part, may be zero, Lg, int
c = Coded 's','d','f' for scalar, diagonal, full, Qtype, char
d = Lags in ARCH (moving average) part, may be zero, Lr, int
e = Coded 's','d','f' for scalar, diagonal, full, Ptype, char
f = Lags in leverage effect of GARCH, may be zero, Lv, int
g = Coded 's','d','f' for scalar, diagonal, full, Vtype, char
h = Lags in additive level effect, may be zero, Lw, int
i = Coded 's','d','f' for scalar, diagonal, full, Wtype, char
j = Lags in polynomial part, Lp, int
k = Degree of the polynomial in z, Kz, int
l = Degree of interactions in z, Iz, int
m = Max degree of z polynomial that depends on x, maxKz, int
n = Max interaction of z polynomial that depends on x, maxIz, int
o = Degree of the polynomial in x, Kx, int
p = Degree of the interactions x, Ix, int

